'*********************************************************
'*                                                       *
'*   ZiCog_cpm v1.0_rc_5.5                               *
'*                                                       *
'*   A simulation of a Zilog Z80 based computer system   *
'*   Capable of running CP/M (v2.2 and 3)                *
'*                                                       *
'*                                                       *
'*   (C) 2009 Michael Rychlik                            *
'*                                                       *
'*********************************************************
'
'v0.0  -  1 Mar 2009 - Original version
'
'v0.1  -  4 Mar 2009 -
'
'v0.2  -  7 Mar 2009 -
'
'v0.3  -  7 Mar 2009 -
'
'v0.4  - 16 Mar 2009 - Replaced T8080 diagnostic with ZICOGTST.COM. This is a version of EXZ80DOC.COM from
'                      the SIMH AltairZ80 simulator with undocumented instructions removed and tests rearranged
'                      in execution time order (almost). T8080.COM no longer passes tests because due to the
'                      implementation of Z80 flag settings rather than 8080.
'                      Fixed up SIMH I/O simulation for console, disks on SD card etc.
'
'v0.5  - 18 Mar 2009 -
'
'v0.6  - 30 Mar 2009 - Removed "outa" capability in output emulation.
'
'v0.7  - 13 Mar 2009 - Set to use zicog version 0.7.
'
'v0.8  -  3 May 2009 - Added SIMH AltairZ80 style hard disk emulation.
'                      Conformant with SIMH ALTAIR 8800, BIOS V1.25, 1 HD, 19-Apr-2009
'
'v0.9  -  5 May 2009 - Added conditional compilation to select Z80 or 8080 emulation. (#ifdef CPU_Z80/CPU_8080)
'                    - Added PropDemoBoard as a conditional compilation option and checks for missing definitions.
'                    - Moved configuration #defines to zicog_demo so they are all in one place.
'                    - Fixed hard disk reset command.
'                    - Added DPB definition for 720K Super I/O Floppy, for use as RAM disk.
'                    - Added some sanity checking on selected HDSK drive, track and sectors.
'                    - Added conditional compilation (#ifdef HEARTBEAT) to enable hearbeat LED flashing from Z80 emulator
'
'v0.10 - 19 Aug 2009 - Incorporate all Cluso's and DrAcula's changes
'                      from zicog_demo009_rr097.spin as listed below.
'                    - Removed keyboard object and all references to kb.something
'                    - Removed TV object and all references to TV.something
'                    - Replaced kb and TV with UART using FullDuplexSerial
'                    - Removed FrequencyCount object and all reference to fcounter (no one was using it)
'                    - Removed a "repeat" from in_hdskport after "HDSK ERROR 5", saves the survey program hanging.
'                    - Changed in_condata to use UART.rxcheck, saves the survey program from hanging.
'                    - Added disk I: to the disk start up message.
'                    - It is now necessary to use SPACE to single step instructions.
'
'1.0_rc_2 - 28 Aug 2009 - Changed to use 4 sectors of 128 bytes per SD block instead of just 1.
'                    - 1 SD block is held in a "cache" at any time so that reads from sectors
'                    - already in the cache do not need to access the SD. Writes to sectors currently in
'                    - the cache do not need to be preceded by sector fetches.
'
'1.0_rc_3 - 5 Sep 2009 - Moved definition of the Z80 RAM area (ram_z80) from zicog_demo.spin
'                      to zicog.spin and placed the Z80 emulation PASM code inside it!.
'                      This save ~2K HUB RAM as the space occupied by the emulator COG code
'                      is now recycled as Z80 RAM when the COG is up and running.
'                      This results in being able to run 24K CP/M system on a Prop Demo board again
'                      which in turn allows booting from disk again.
'                    - Added disk parameters for 440K RAM disk as dpb_3.     <==================???? dpb_6 ???
'                    - Removed all code for handling Altair 1M floppy disks.
'                      Brutal I know but they were messy to use with SD cards and very slow.
'                      If there are anyone really wants them back we can dig the code out of the archive
'                    - File renamed zicog_cpm.spin (from zicog_demo.spin)
'                    - Removed old cpu_control cpu parameter
'                    - Removed ram and rom junk from cpu parameters
'1.0_rc_4 - 7 Sep 2009 -
'
'1.0_rc_5 - 8 Sep 2009 - Made at least compilable for TriBlade and CPU_Z80
'                      - Changed sd_block to sd_block_numer throughout TriBlade code.
'                      - Added warning if no platform define is specified.
'
'1.0_rc_5.5-8 Sep 2009 - Added LoadZ80Ram to read disk boot sectors to Z80 RAM
'                        prior to start.
'                        Requires new BOOTGEN to lay down those sectors.
'
'' RR20090404   Fix required sections for compiling by homespun024.exe
''              sdspiqasm, PC_Keyboard, PC_Text, FrequencyCount all reguired org -> org 0
''              zicog: jmp requires # for line 374, and a description for jmp without # to say jmp/ret(indirect) 
''
'' +-------------------------------------------------------------------------------------------------------+
'' | Modifications to ZiCog v0.5 for use on Cluso's TriBladeProp PCB (Blade #2)                       vxxx |
'' +-------------------------------------------------------------------------------------------------------+
'' |  Author:            "Cluso99" (Ray Rodrick)                                                           |
'' |  Copyright (c) 2009 "Cluso99" (Ray Rodrick)                                                           |
'' |  License            Modifications only under MIT License                                              |
'' +-------------------------------------------------------------------------------------------------------+
'' RR20090317   Cluso99: Modifications to ZiCog v0.5 for use on the TriBladeProp PCB (Blade #2)
'' RR20090317   start to use TriBladeProp & SRAM                             
''              Ram is only $4F00 - shouldn't it be $5000 ??                 
''              Remove heartbeat and FrequencyCounter.spin and fcounter.xxx calls
''              Remove sdspiqasm.spin and disk.xxx calls                     
''              Remove output LEDs                                           
'' RR20090318   add sram_start (ram offset)
''              Rom is $0180 long beginning $FF00 - shouldn't it be $0100 long ??
'' RR20090318   move from v0.4 to v0.5
'' RR20090318   _rr004 version
'' RR20090319   make 64KB CPM version
''              change msize 20 -> 64; disable free_ reservation in hub
'' RR20090320   _rr006 version: add SRAM disk
''              _rr007 complete and enable SRAM disk
'' RR20090321   _rr008 use new CBIOSX_triblade.COM
''              preload SRAM disk from cpm2.dsk using cpm2SD_sram_rr008 (truncates)
''              enabled readblock & writeblock (no clear on start). DIR -> "no file"
'' RR20090321   _rr009 incorporate cpm2SD_sram_rr008 to preload SRAM with cpm disk
''              short of stack space - disable part of print_regs debug code
''              _rr010 DIR says "no file"
''              _rr011 use sdsram_rr011 for loading xxxx.COM files to Z80RAM (64KB)
''              _rr015 preload SRAM disk first using cpm2SD_debug4 & SRAM_driver_015
'' RR20090331   _rr021 use TriBladeProp_Blade2_201.spin driver to load Z80 Ram and CPM Ram Disk
''              _rr022 bugfix (sector 127->128) Use Cluso's VB6 to load Ram from PC
''              _rr023 works loading SRAM and verified (using VB) for Z80 & Ram Disk 448_512KB
''                     gives A> prompt but DIR still does not work
''              _rr024 LoadSRAM optional constant
''              _rr025 the Ram Disk files have 137 byte sectors (not 128 as first thought) so fix in spin for now
''                     Hooray it does DIR :-)
''                     Warning, likely to require more stack space before continuing !!!
'' RR20090401   _rr026 start to load from microSD the SRAM 1MB "DRIVE_1M.RAM" (uses first 128 bytes of 512 byte sectors)
''              _rr027 remove the cpm code included files (z80ram tpa, ccp, etc) to freeup space
'' RR20090402   _rr028 add SD_Trainer_rr008 "FullDuplexSerialPlus" & "fsrwFemto" for microSD load to SRAM
''              _rr034 writing issue - try new driver _v202 (not the problem)
'' RR20090403   _rr035 (proved sram walking code ok, so some interaction with sd card drivers?)
''              _rr036 Fixed bug: SD card not tristating D0 - fixed sdspiFemto_rr001.spin
''              _rr040 Loading 1MB SRAM from "DRIVE_1M.RAM" on microSD and getting A>
''              _rr041 Does a DIR listing (SRAM loaded, so loading commented out)
'' RR20090404   _rr043 Migrate to ZiCog v0.006 & homespun compiler
'' ---------------------------------------------------------------------------------------------------------
'' RR20090404   _rr044 Migrate to ZiCog v0.006 & homespun compiler
''              _rr045 question to load SRAM from microSD, Fix ram_size
''              _rr045 Write working on _rr042(later), added here & works
''                     Note does not work with zicog_rr044 but does with zicog005_rr021
'' RR20090405   _rr046 Remove clear screen for loading from PropCMD - working & posted
''              _rr047 remove special prop UTF-16 characters
''              _rr048 done
'' RR20090407   _rr049 modify to use direct sector address for disk files on microSD card
''              _rr050 loading the 64KB+960KB using direct SD sector addressing
'' RR20090408   _rr051 convert keys
''              _rr052 slow to 9600 baud and do comms between Blades #2<-->#1
''              _rr053         4800
'' ---------------------------------------------------------------------------------------------------------
'' SKIP _rr052 & _rr053 (for now)
'' RR20090416    _rr055 (from _rr051) clear sram 0-64KB and only load $FF00-$FFFF DSKBOOT.COM
''                      now CPM should boot from the sram disk - does not work, so place jmp to $FF00 at $0000
'' RR20090416    _rr056 working (removed jmp $FF00 at $0000)
''               _rr057 get 2nd disk running (on SD card)
'' RR20090417    _rr058 continue
''               _rr059 continue
''               _rr060 continue (new SD files DSK_CPM2.DSK & DSK_CPM3.DSK) older formats incorrect
''                      loads ok but will not take kbd input due to restarts???
''               _rr061 continue (add #define sddisk)
''               _rr062 conflict (add #define sduseinit) prevent readsd from rebooting cog each time - works
''               _rr063 conflict (#define sddisk) - no kbd input at A> - now works, but resetting tbp2 cog !!
''                      but not using the sd disk data yet
''               _rr064 use external CPM2 disk on sd card (#define sdCPM2)
''               _rr065 (#define RamDisk) Note to fix resetting tbp2 from resetting when accessing sdcard
''                      working
''               _rr066 tidy code (tbp2 driver requires "idle" command instead of start/stop)
''                      add drive1base and 2 disks   - works
'' RR20090420    _rr067 try to do syscpm2 - works
'' RR20090421    _rr068 add drives C...H
'' RR20090421    _rr069 try cpm3
''                      try heaters new nonbanked cpm3 - works but have to enter 2 characters to get 1 input !!
'' RR20090429    _rr070 fixed tbp2 idle/active for bus conflict "TriBladeProp_Blade2_202"
''               _rr072 add #define CPM3 (not tested)
'' RR20090430    _rr073 new nonbanksleep CMP3 version  (doesn't fix keyb bug)
''                      try and use ZiCog v0.07_rr070
'' RR20090504    _rr074 try ZiCog v0.08_rr073 (_rr080 working)
'' ---------------------------------------------------------------------------------------------------------
'' RR20090504    _rr080 combine v0.08 demo with demo_rr074
''                      CPM3 still has kbd bug
''               _rr081 add #define hdskenable
'' ---------------------------------------------------------------------------------------------------------
'' RR20090507    _rr090 combine v0.09
'' RR20090805    _rr091 use standalone in Blade #2 (released 20090805)
'' RR20090806    _rr092 use PC_Text & PC_Keyboard (replace FDXplus) 
''               _rr093 add #define addLF
'' RR20090817    _rr094 Dr_Acula: N8VEM & fix dropping received characters (_rr092_james2)
'' RR20090818    _rr095 fix reversed rxavail
''               _rr096 add 4 x Hard Disk support (I:,J:,K:,L:)  partial only
'' Dr20090818    _rr097 Dr_Acula: modifications for CRLF (_rr096_j18)
'' ---------------------------------------------------------------------------------------------------------
'' MR20090819    v0.10  modifications by heater (see above)
'' ---------------------------------------------------------------------------------------------------------
'' RR20090819    _rr100 add #define fix137 (for the fudge 137 byte sectors for SIMH); minor code tidyup
'' RR20090819    _rr101 remove fix137/preset_137 not required ($A7,$0E,$D2.....$B7,$C8,$CD,$A5,$2B,$00)
'' RR20090820    _rr102 many thanks to Dr_Acula and Heater for the combined testing evening.
''                      fix hard disk on tbp
''                      force jmp to $FF00 (faster)
''                      some tidyup
'' RR20090821    _rr103 allow 26 drive(SD file offset) parameters
''                      R: is RamDisk
'' RR20090823    _rr104 Note 16 drives is the max for CPM, SIMH has 12 (A:..H: + I:..L:)
''                       so allow 8x1MB Floppy + 4x HardDisk (2x8MB, 1x..KB Ram, 1x..MB large)
'' RR20090829    _rr105 Use "DSKBOOT_.ROM" to boot (load into $FF00)
''                      Use HDRIVE_I.DSK & HDRIVE_J.DSK filled with "$E5" and then PIP I:=A:*.*
''               _rr106 Setup RamDisk F: with "$E5" - works
''               _rr107 Make RamDisk hard disk J: - works BUT bounds checking invalid (see STAT)
''                      General code tidy, change to common CheckError routine and check <0
''                      Note: hdp block size mods do not work
'' ---------------------------------------------------------------------------------------------------------
'' MR20090907  v1.0_rc4 modifications by heater (see above)
''                      rename ZiCog_demo -> ZiCog_cpm
'' ---------------------------------------------------------------------------------------------------------
'' RR20090907    _rr108 add heaters mods (fails compilation)
'' RR20090908    _rr109 try make it compile    
'' ---------------------------------------------------------------------------------------------------------
'' 1.0_rc_5.5b-8 Sep 2009  modifications by heater (see above)
'' RR20090913    _rr112 mod to pass all SD pins to sdspiqasm_rr001
'' RR20090913    -rr113 now runs 25K CPM2.2 version on TriBlade hardware WITHOUT USING SRAM.


CON

  baud          = 115_200                               '\ Uncomment one for your serial speed
' baud          = 38_400                                '/ Uncomment one for your serial speed


'USER CONFIGURATION OPTIONS

'Uncomment one of the following CPU emulation types.
'#define CPU_Z80
#define CPU_8080

'Uncomment one of the following platform types.
'#define TriBladeProp
#define PropDemoBoard
#define UseTriBladeSD                                   'Uncomment to run PropDemoBoard on TriBlade hardware
#define Boot_from_SD
'#define Boot_from_SD_SPIN


'#define banked_memory                                  'Uncomment if banked memory is emulated.(48K+16K)+4x48KB=256KB
'#define SingleStep                                     'Uncomment if "single step" emulation required.
'#define HEARTBEAT                                      'Uncomment if Z80 emulation is to flash a heartbeat LED
#define Show_Hex                                        'Uncomment to display in hex, else display char


#ifdef TriBladeProp
#elseifdef PropDemoBoard
#else
#warn No HW platform type defined
#endif



#ifdef TriBladeProp
#define diskboot                                        'boot using cleared ram and $FF00 dskboot.com
#define RamHDisk                                        'use Ram Disk for HardDrive J:<==========???
#define cpm2                                            'Uncomment for CPM2.2 else CPM3
#endif



#ifdef TriBladeProp
  _clkmode      = xtal1 + pll16x
  _xinfreq      = 5_000_000

' microSD pins on TriBladeProp Blade #2                  
  spiDO         = 9
  spiClk        = 28
  spiDI         = 8
  spiCS         = 14

  cpmsize       = 64   * 1024                           'CPM Ram size (see msize below)
  sramsize      = 512  * 1024                           'Uncomment for 512KB SRAM
'  sramsize      = 1024 * 1024                           'Uncomment for 1MB  SRAM


#else

#ifdef UseTriBladeSD
  _clkmode      = xtal1 + pll16x
  _xinfreq      = 5_000_000
' microSD pins on TriBladeProp Blade #2
  spiDO         = 9
  spiClk        = 28
  spiDI         = 8
  spiCS         = 14
#else
  _clkmode      = xtal1 + pll8x
  _xinfreq      = 10_000_000
' microSD pins on PropDemoBoard
  spiDO         = 0
  spiClk        = 1
  spiDI         = 2
  spiCS         = 3
#endif

#endif

OBJ

  cpu      : "zicog"                                    'Zilog Z80 emulator in one COG
  UART     : "FullDuplexSerial_rr004"                   'FDX v1.2 with mods
#ifdef TriBladeProp
  tbp2     : "TriBladeProp_Blade2_202"                  'Blade2 hardware driver
  sd       : "fsrwFemto"                                'SD Software used in FemtoBASIC
#else
  sd       : "sdspiqasm_rr001"                          'Low level access to SD card (passes all pins)
#endif


CON

'Z80 Program Status Word bits
sign_bit        = %10000000
zero_bit        = %01000000
aux_bit         = %00010000
parity_bit      = %00000100
carry_bit       = %00000001


'I/O control block commands
io_cmd_out      = $01
io_cmd_in       = $02
io_cmd_break    = $03

'
'       A read to the data port gets the buffered character, a write
'       to the data port writes the character to the device.
constat         = 16                    ' sio port 1 status port
condata         = 17                    ' sio port 1 data port
punstat         = 18                    ' sio port 2 status port
pundata         = 19                    ' sio port 2 data port

simhport        = $FE
srport          = $FF                  ' Altair front panel switch register port

ledport         = $00

' UART constants for N8VEM
n8vem_uart_test = $6D
n8vem_uart_in   = $68
n8vem_uart_out  = $68

' masks for SIO controller (constat, punstat)
mout            = $02                   ' output allowed mask

' commands for SIO controller (constat, punstat)
creset          = 3                     ' reset command

' constants for hard disk port
nhdisks         = 8                     'The number of supported hard disks

hdskNull        = 0                     ' No command
hdskReset       = 1                     ' command to reset controller
hdskRead        = 2                     ' read command
hdskWrite       = 3                     ' write command
hdskParam       = 4                     ' Get HD parameters command
hdskport        = $fd                   ' control port for simulated hard disk

#ifdef UseTriBladeSD
hdsk_base_block = $202c7                ' First block of first hard disk in SD "zipm2.dsk"
#else
hdsk_base_block = 0                     ' First block of first hard disk in SD.
#endif

'constants for SIMH port
simhreset       = 14
stoptimer       = 22
setbankselect   = 12
hasbankedmemory = 18

DAT

'The Z80 Registers.
'Do not change the order of these.
'Registers are defined in pairs as words so as to enable them to be
'accessed individually as BYTE or in pairs as WORD.
'The ordering makes it easier to find a register specified by a 3-bit
'register field within an op code or to swap between main and
'alternate register sets. 
register_file word
bc_reg        word      0
de_reg        word      0
hl_reg        word      0
af_reg        word      0
bc_reg_alt    word      0 
de_reg_alt    word      0
hl_reg_alt    word      0
af_reg_alt    word      0
im_reg        word      0
ix_reg        word      0
iy_reg        word      0
sp_reg        word      0
pc_reg        word      0
  
'IO interface area. Do not change the order of these.
io_cmd_block  long
io_command    long      0
io_port       long      0            
io_data       long      0


'Hard Disk Parameter Blocks
'Each DPB is 17 bytes as defined for DPBs in CP/M followed by 2 bytes sector size as used by SIMH
hdsk_dpb
dpb_0                           '8MByte Default SIMH Altair HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $05     'data allocation Block SHift factor
:blm          BYTE      $1F     'data allocation block mask
:exm          BYTE      $01     'EXtent Mask
:dsm_low      BYTE      $F9     'maximum data block number (low_byte)
:dsm_high     BYTE      $07     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $03     'total number of directory entries (high byte)
:al0          BYTE      $FF     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)
                                'N.B. SS must be 128 for CP/M 2 can be varied for CP/M 3 hard disks.

dpb_1                           '8MByte Default SIMH Altair HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $05     'data allocation Block SHift factor
:blm          BYTE      $1F     'data allocation block mask
:exm          BYTE      $01     'EXtent Mask
:dsm_low      BYTE      $F9     'maximum data block number (low_byte)
:dsm_high     BYTE      $07     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $03     'total number of directory entries (high byte)
:al0          BYTE      $FF     'determine reserved directory blocks
:al1          BYTE      $00     'determdirine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)
                                'N.B. SS must be 128 for CP/M 2 can be varied for CP/M 3 hard disks.

dpb_2                           '8MByte Default SIMH Altair HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $05     'data allocation Block SHift factor
:blm          BYTE      $1F     'data allocation block mask
:exm          BYTE      $01     'EXtent Mask
:dsm_low      BYTE      $F9     'maximum data block number (low_byte)
:dsm_high     BYTE      $07     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $03     'total number of directory entries (high byte)
:al0          BYTE      $FF     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)

dpb_3                           '8MByte Default SIMH Altair HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $05     'data allocation Block SHift factor
:blm          BYTE      $1F     'data allocation block mask
:exm          BYTE      $01     'EXtent Mask
:dsm_low      BYTE      $F9     'maximum data block number (low_byte)
:dsm_high     BYTE      $07     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $03     'total number of directory entries (high byte)
:al0          BYTE      $FF     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)

dpb_4                           '8MByte Default SIMH Altair HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $05     'data allocation Block SHift factor
:blm          BYTE      $1F     'data allocation block mask
:exm          BYTE      $01     'EXtent Mask
:dsm_low      BYTE      $F9     'maximum data block number (low_byte)
:dsm_high     BYTE      $07     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $03     'total number of directory entries (high byte)
:al0          BYTE      $FF     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)

dpb_5                           '8MByte Default SIMH Altair HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $05     'data allocation Block SHift factor
:blm          BYTE      $1F     'data allocation block mask
:exm          BYTE      $01     'EXtent Mask
:dsm_low      BYTE      $F9     'maximum data block number (low_byte)
:dsm_high     BYTE      $07     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $03     'total number of directory entries (high byte)
:al0          BYTE      $FF     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)

dpb_6                           '1MByte Altair mits2 floppy as HDSK params (no skew)
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $04     'data allocation Block SHift factor
:blm          BYTE      $0F     'data allocation block mask
:exm          BYTE      $00     'EXtent Mask
:dsm_low      BYTE      $EF     'maximum data block number (low_byte)
:dsm_high     BYTE      $01     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $00     'total number of directory entries (high byte)
:al0          BYTE      $F0     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)
                                'N.B. SS must be 128 for CP/M 2 can be varied for CP/M 3 hard disks.

dpb_7                           '440K RAM disk as HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $04     'data allocation Block SHift factor
:blm          BYTE      $0F     'data allocation block mask
:exm          BYTE      $00     'EXtent Mask
:dsm_low      BYTE      $DF     'maximum data block number (low_byte)  $006F=111*32*128=454,656=444KB <======??????
:dsm_high     BYTE      $00     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $00     'total number of directory entries (high byte)
:al0          BYTE      $F0     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $00     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)
                                'N.B. SS must be 128 for CP/M 2 can be varied for CP/M 3 hard disks.


VAR     

'Parameter list for CPU objects, do not change order.
long reg_base                   'Address (in HUB)of Z80 register file
long io_base                    'Address (in prop RAM) of the I/O contol interface

word err                        'Error code return value

long switch_reg 
long simhcmd

'''???long dbgcnt
'''???long firsttime
'''???long sector

#ifdef TriBladeProp
'For SRAM CPM Disk Loading
  long  filename
  long  srcname
  byte  buffer[16]
  long  sectorsize
'For microSD routines

  long  ioControl[2]               ' SD parameters                     
  long  drive_base[18]             ' SD base address of drives 0..15 (A:..L:) & ram,ram,rom sector 0
  long  blockno                    ' blockno within file
  long  sdsector                   ' sector no 
  long  count                      ' count of chars to read
  byte  buff[8192]                 ' SD buffer
#endif


  byte hdsk_reset_count            'Counts number of consequiteve hdsk reset commands.
  byte hdsk_current_command        'The hdsk command currently in progress
  byte hdsk_last_command           'The last complete hdsk command issued
  byte hdsk_command_pos            'Counts through the outputed command parameter bytes (OUTs)
  byte hdsk_reply_pos              'Counts hdisk port reply bytes (INs)
  long hdsk_dpb_ptr                'Pointer into DPB tables.
  byte hdsk_drive                  'The hdsk drive to read/write with this command
  byte hdsk_sector                 'The hdsk sector to read/write with this command
  word hdsk_track                  'The hdsk track to read/write with this command
  word hdsk_dma                    'The DMA address to read/write from with this command

  long sector_number               'Sector of a disk starting from 0
  long sd_block_number             'SD card 512 byte block number starting from 0
  long block_offset                'Byte offset of a sector within a block
  long current_block_number        'The SD block currently in the disk buffer
  byte disk_buff[512]              'The disk I/O buffer holds one SD card block.


PUB start

  waitcnt(clkfreq*3 + cnt)      'delay (so you can get PropTerminal active)     <=== 3 secs

  current_block_number := -1    'No SD block read into buffer yet so set invalid value

  'Start the terminal
  UART.start(31,30,0,baud)      'start the uart Tx and Rx and set the baud rate

#ifdef TriBladeProp
  UART.str(string("ZiCog 1.0_rc_5.5b on the TriBladeProp v0.113"))
  crlf
'  UART.str(string("8 bit, parity=None, 1 stop bit, Flow control=None"))
'  crlf
#else
  UART.str(string("ZiCog 1.0_rc_5.5b on the Prop Demo Board"))
#ifdef UseTriBladeSD
  UART.str(string(" TBP v0.113"))
#endif
  crlf
#endif

  reg_base        := @register_file
  io_base         := @io_cmd_block
  switch_reg      := $08                                '$08 is the switch setting SIMH recommends for 4k Basic etc to run
'''???  firsttime       := 0                                  'if 0 display first disk sector read by Z80

#ifdef TriBladeProp

'------------------------------------------------------------------------------
' Start the disks (from SD card)
' Load SRAM with Z80 RAM (64KB) & optional CPM Ram Disk (448KB/960KB)
'------------------------------------------------------------------------------
  UART.str (string("Starting Hdisks... "))
  crlf
  blockno := 0
  count := 8192
  FindSDblock                                           'set the SD card file/drive base(s)

  UART.str(string("Loading SRAM... "))
  crlf
  LoadSRAM                                              'clear 64KB sram, diskboot at $FF00, fill >64KB with $E5

#elseifdef PropDemoBoard

'------------------------------------------------------------------------------
' Start the disks (from SD card)
'------------------------------------------------------------------------------
  UART.str (string("Starting Hdisks (A:B:C:D:E:F:G:H) ... "))
  crlf
'  err := \sd.start(0)
  err := \sd.start(spiDO,spiClk,spiDI,spiCS)            'set SD card pins
  DisplayError(err)

#endif

#ifdef BOOT_FROM_SD_SPIN
  UART.str(string("Loading Z80 ram..."))
  LoadZ80Ram
#endif

'------------------------------------------------------------------------------
' ZiCog Emulation starts here...
'------------------------------------------------------------------------------
#ifdef CPU_Z80
  UART.str(string("Starting Z80 emulation..."))
#else
  UART.str(string("Starting 8080 emulation..."))
#endif
  crlf
  err := \cpu.start(@reg_base)
  DisplayError(not err)

  'Start the Altair I/O simulation
  service_io

'------------------------------------------------------------------------------
'
'------------------------------------------------------------------------------



'==============================================================================
' TriBladeProp Routines
'==============================================================================
#ifdef TriBladeProp

PRI FindSDblock | i, n, r                               'read blockno of file for count into buffer

  r := sd.start(@ioControl)                             'start the SD routines
  CheckError(not r)
  r := sd.mount(spiDO,spiClk,spiDI,spiCS)               'mount the SD
  CheckError(r)


  'HDisks A:,B:,C:,D:,E:,F:,G:,H:
#ifdef cpm2
  FindDriveBase(0, string("HDCPM2_A.DSK"))              'set the SD card drive base(s)  8MB CPM2.2
#else
  FindDriveBase(0, string("HDCPM3_A.DSK"))              'set the SD card drive base(s)  8MB CPM3 nonbanked
#endif
  FindDriveBase(1, string("HDRIVE_B.DSK"))              'set the SD card drive base(s)  8MB
  FindDriveBase(2, string("HDRIVE_C.DSK"))              'set the SD card drive base(s)  8MB
  FindDriveBase(3, string("HDRIVE_D.DSK"))              'set the SD card drive base(s)  8MB
  FindDriveBase(4, string("HDRIVE_E.DSK"))              'set the SD card drive base(s)  8MB
  FindDriveBase(5, string("HDRIVE_F.DSK"))              'set the SD card drive base(s)  8MB
  FindDriveBase(6, string("HDRIVE_G.DSK"))              'set the SD card drive base(s)  1MB equiv floppy
  FindDriveBase(7, string("HDRIVE_H.DSK"))              'set the SD card drive base(s)  <1MB RamDisk

  'other files Q:,R:
  FindDriveBase(16,string("DSKBOOT_.ROM"))              'set the SD card drive base(s)    $FF00 for boot
  FindDriveBase(17,string("ZIBOOT  .COM"))              'set the SD card drive base(s)    $FF00 for boot

  r := sd.stopSDCard                                    'stop SD card, but do not stop cog
  CheckError(r)

PRI FindDriveBase(n, st) | r                            'set the SD drive_base for the drive no
  UART.tx($41 + n)
  UART.tx(":")                                          'A:...
  UART.str(st)

  r := sd.popen(st,"r")                                 'Open file for read (blank image)
  CheckError(r)
  r := sd.pread(@buff,1)                                'get a char to find the sector address ****
  CheckError(r)
  drive_base[n] := ioControl[0] - 1                     'get the sectorno of the file (returns the sector+1)

#ifdef RamHDisk
  if n == 7                                             'H: is the RamDisk
    UART.str(string(" <-RamDisk"))
#endif
  if n == 16
    UART.str(string(" <-bootcode"))
  if n == 17
    UART.str(string(" <-bootcode??"))
  crlf


PRI LoadSRAM | i, r, b1, b2

  count := 8192                                         'use 8KB blocks
  b1 := cpmsize / count                                 '8KB blocks in 64KB CPM RAM
  b2 := sramsize / count                                '8KB blocks in 512KB/1MB SRAM


'clear the whole 64KB sram first (need a command in tbp2 to do this)
  r := tbp2.start                                       'start the triblade2 driver
  CheckError(r)

  repeat i from 0 to count - 1                          'set 8KB block = $00
    buff[i] := 0

  repeat i from 0 to b1 -1                              '0 to 7: write 8 x 8KB blocks =$00 to sram
    r := tbp2.DoCmd("W", @buff, i * count, count)       'write to sram
    CheckError(r)

  buff[0] := $C3                                        'force jmp to $FF00 at $0000 (faster)
  buff[1] := $00
  buff[2] := $FF
  r := tbp2.DoCmd("W", @buff, 0, 3)                     'write to sram
  CheckError(r)

  r := tbp2.idle                                        'give up the bus
  CheckError(r)

'read the DSKBOOT_.ROM file (256 bytes)
  r := sd.initSDCard(spiDO,spiClk,spiDI,spiCS)          'init SD card, but do not restart cog
  CheckError(r)

  r := sd.readSDCard(drive_base[16], @buff, 256)        'get 256 bytes   "DSKBOOT_.ROM"
'?  r := sd.readSDCard(drive_base[17], @buff, 256)        'get 256 bytes   "ZIBOOT  .COM"
  CheckError(r)

  r := sd.stopSDCard                                    'stop SD card, but do not stop cog
  CheckError(r)

'copy the 256 bytes to $FF00-$FFFF in sram
  r := tbp2.active                                      'activate the bus
  CheckError(r)

  r := tbp2.DoCmd("W", @buff, $FF00, 256)               'write to sram
  CheckError(r)

'now set the buffer to "$E5" and fill the remainder of sram for use as empty RamDisk
  repeat i from 0 to count - 1                          'set 8KB block = $E5
    buff[i] := $E5

  repeat i from b1 to b2 - 1                            '8 to 63/127: write 56/120 x 8KB blocks of sram above first 64KB
    r := tbp2.DoCmd("W", @buff, i * count, count)       'write to sram
    CheckError(r)

' ^ leave the tbp2 bus active (U23 sram selected and latched)

PRI show(s) | i, addrx
  addrx := blockno * count
  UART.hex(addrx,6)
  UART.tx(" ")
  repeat i from 0 to 15                                 'only display first xx chars per block
#ifdef Show_Hex
    UART.hex(buff[i],2)
    UART.tx(" ")
#else
    UART.tx(buff[i])
#endif
  crlf


PRI CheckError(r)
  if r < 0
    UART.str(string("***Error*** $"))
    UART.hex(r,2)
    UART.str(string(" ***HALTED***"))
    crlf
    sd.stop
    tbp2.stop
    repeat                                              '<==============  LOOP HERE

'------------------------------------------------------------------------------
#endif
'==============================================================================

PRI DisplayError(errx)
  if errx
    UART.str(string(" ***Failed***"))
  else
    UART.str(string("Passed, please wait..."))
  crlf


#ifdef BOOT_FROM_SD_SPIN
PRI LoadZ80Ram | bootBlock, sect, z80_address
' Reads 88 sectors (22 SD blocks) from SD to Z80 RAM.
' FIXME: Should allow selectable boot disks
  block_offset := 0
  z80_address := 0
  repeat bootBlock from 2 to 23
#ifdef PropDemoBoard
    err := \sd.readblock(bootBlock, @disk_buff)
#elseifdef TriBladeProp
    'Cluso something here perhaps
#endif
    if err
      UART.str(string("Error hdisk boot read"))
      return
    repeat sect from 0 to 3
      block_offset := sect * 128
      cpu.dma_write (z80_address, (@disk_buff) + block_offset)
      z80_address += 128
#endif


'------------------------------------------------------------------------------
' Service I/O
'------------------------------------------------------------------------------
PRI service_io
'Altair I/O simulation service loop
  repeat
    case io_command
      io_cmd_break:                                     'CPU break
        on_break    
      io_cmd_in:                                        'An input operation
        on_input
      io_cmd_out:                                       'An output operation
        on_output

PRI on_break
'Handle CPU break condition
'No, this never existed in an Altair (or SIMH simulator)
  print_regs
  'waitcnt (cnt + 80_000)
  repeat while UART.rx <> $20                           'wait for 'space'
  io_command := 0


PRI on_input
'Handle Z80 IO port reads
  case io_port
    srport:                                             'Altair front panel switches register
      in_srport
    simhport:                                           'SIMH control port
      in_simhport          
    constat:                                            'Console status port.
      in_constat
    condata:                                            'Console data port
      in_condata
    n8vem_uart_test:                                    ' test port $6D
      in_test_n8vem                                     ' is a byte ready returns 0 for no byte, 1 for a byte
    n8vem_uart_in:                                      ' ask for input from port $68
      in_uart_n8vem                                     ' get the byte
    hdskport:
      in_hdskport                                       'SIMH style hard disk port
    other:
      'UART.str(string("Read from unimplemented I/O Port &H"))
      'UART.hex(io_port,2)
      'crlf
      io_data := 0
  io_command := 0                                       'Set flag ready for next io

PRI in_test_n8vem
'  for xmodem tests if any input from the serial port
' input present = bit 0. Output ready to go = bit 5 (always high). bit1=0 (carrier detect)
  if UART.rxavail
    io_data :=%00100001                                 'yes
  else
    io_data :=%00100000                                 'no

PRI in_uart_n8vem
  if UART.rxavail 
    io_data := UART.rx                                  ' get the byte
  else
    io_data := "?"                                      '<=== at least show a character to indicate a problem

PRI in_srport
'Handle reads from the Altair front panel switch port
  io_data := switch_reg
  UART.str(string("SR port read"))
  crlf

PRI in_simhport
'Handle reads from the SIMH simulator status port
  if simhcmd == hasbankedmemory
    io_data :=  1
    simhcmd := 0

PRI in_constat
'Handle reads from the console status port
  if UART.rxavail                                       'Is there a character in the buffer
    io_data := %00000011                                'Return flag to say a character is available
  else
    io_data := %00000010                                'No: Return no char ready for input

PRI in_condata
'Handle reads from the console data port (N.B. No wait)
  io_data := UART.rxcheck                               'Yes: Return a character (or junk if none available)

PRI hdsk_check_dts : status
'Check that selected hdsk drive, sector and track are within limits
  if (hdsk_drive < nhdisks) and (hdsk_track =< 2048) and (hdsk_sector =< 128)
    status := $00
  else
    UART.str(string("Ilegal hdsk access"))
    DisplayHDiskDTS
    status := $FF


PRI refresh_hd_cache
'Ensure that the sector number in use is in the current SD block in memory.
'If not load that block to disk buff from SD card.
'

'<=======================triblade needs a fix here !!!

  'WARNING This is jumping over 8MB chunks of SD card.
  sector_number := hdsk_drive * (32 * 2048)           'Skip on to the first sector of the selected drive
  sector_number += (hdsk_track * 32)                  'Find correct block for the selected track..
  sector_number += hdsk_sector                        '..and sector
  sd_block_number := sector_number >> 2               'Determine the SD block of this sector..
  sd_block_number += hdsk_base_block                  '..starting at first block of first HD is SD.
  block_offset := (sector_number & %11) * 128         'Determine the position of this sector within the block
  if sd_block_number <> current_block_number          'If the block is not in disk buffer then load it
#ifdef PropDemoBoard
    err := \sd.readblock(sd_block_number, @disk_buff + block_offset) 'Read the block !
#elseifdef TriBladeProp
    'Cluso something here perhaps
#endif
    if err
      UART.str(string("Error hdisk cache read"))
      DisplayError(err)
      DisplayHDiskDTS
      io_data := $FF                                  'Return error code
    else
      current_block_number := sd_block_number         'Make this new block current


PRI in_hdskport_read | r
'Handle read from hdskport during a disk read command
'This read triggers the actual disk sector transfer and must return an error code (To the Z80 not the calling Spin)
'N.B. sectors are mapped to blocks 4 to 1

'  UART.str(string("hdisk read "))
  if hdsk_check_dts                                   'Check drive, track and sector are sane
    io_data := $FF                                    'Return error if not
    return

'<=======================triblade needs a fix here !!!




'----------------------------------------------------------------------------------------------
#ifdef TriBladeProp
#ifdef RamHDisk
'----------------------------------------------------------------------------------------------
  if hdsk_drive == 1                                    'J: is RamDrive

    sd_block_number := (hdsk_track * 32) + hdsk_sector
    err := tbp2.DoCmd("R", @disk_buff, sd_block_number*128 + cpmsize, 128)

  else

    r := tbp2.idle                                      'give up the bus
    CheckError(r)
     
    r := sd.initSDcard(spiDO,spiClk,spiDI,spiCS)        'init the SD
    CheckError(r)

    ' this method relies on the SD file using contiguous sectors !!! Only the first 128 bytes of 512 bytes used
    sd_block_number := (drive_base[hdsk_drive+8]) + (hdsk_track * 32) + hdsk_sector      '8x floppies, then 4x hdisks
    err := sd.readSDCard(sd_block_number, @disk_buff, 128)     'read a sector
'   CheckError(err)

    r := sd.stopSDcard                                  'stop the SD (free up pins, but keep cog running)
    CheckError(r)
     
    r := tbp2.active                                    'activate the bus
    CheckError(r)
'----------------------------------------------------------------------------------------------
#else
'----------------------------------------------------------------------------------------------
  r := tbp2.idle                                        'give up the bus
  CheckError(r)

  r := sd.initSDcard(spiDO,spiClk,spiDI,spiCS)          'init the SD
  CheckError(r)

  ' this method relies on the SD file using contiguous sectors !!! Only the first 128 bytes of 512 bytes used
  sd_block := (drive_base[hdsk_drive+8]) + (hdsk_track * 32) + hdsk_sector      '8x floppies, then 4x hdisks
  err := sd.readSDCard(sd_block, @disk_buff, 128)       'read a sector
' CheckError(err)

  r := sd.stopSDcard                                    'stop the SD (free up pins, but keep cog running)
  CheckError(r)

  r := tbp2.active                                      'activate the bus
  CheckError(r)
'----------------------------------------------------------------------------------------------
#endif

  if err
    UART.str(string("Error hdisk read "))
    DisplayError(err)
    DisplayHDiskDTS
    io_data := $FF                                      'Return error code
  else
    'write the block out to sram
    r := tbp2.DoCmd("W", @disk_buff, hdsk_dma, 128)     'Virtual "DMA" the sector into DMA buffer address in SRAM
    CheckError(r)
    io_data := $00                                      'Return OK.
'----------------------------------------------------------------------------------------------
#else

'<=======================triblade needs a fix here !!!
  refresh_hd_cache

  'Virtual "DMA" the sector into DMA buffer address in Z80 RAM
  cpu.dma_write(hdsk_dma, @disk_buff + block_offset)
  io_data := $00                                        'Return OK.

#endif
'----------------------------------------------------------------------------------------------
  hdsk_last_command := hdskNull                         'Command completed


PRI in_hdskport_write | r
'Handle read from hdskport during a disk write command
'This read triggers the actual disk sector transfer and must return an error code(To the Z80 not the calling Spin)
'N.B. sectors are mapped to blocks 4 to 1

  if hdsk_check_dts                                     'Check drive, track and sector are sane
    io_data := $FF                                      'Return error if not
    return

'<=======================triblade needs a fix here !!!
  refresh_hd_cache


'----------------------------------------------------------------------------------------------
#ifdef TriBladeProp

#ifdef RamHDisk
'----------------------------------------------------------------------------------------------
  'read the block in from Z80 sram
  r := tbp2.DoCmd("R", @disk_buff, hdsk_dma, 128)       'Virtual "DMA" the sector from DMA buffer address from SRAM
  CheckError(r)

  if hdsk_drive == 1                                    'J: is RamDrive

    sd_block_number := (hdsk_track * 32) + hdsk_sector
    err := tbp2.DoCmd("W", @disk_buff, sd_block_number*128 + cpmsize, 128)

  else

    'read the block in from sram
    r := tbp2.DoCmd("R", @disk_buff, hdsk_dma, 128)     'Virtual "DMA" the sector from DMA buffer address from SRAM
    CheckError(r)

    r := tbp2.idle                                      'give up the bus
    CheckError(r)

    r := sd.initSDcard(spiDO,spiClk,spiDI,spiCS)        'init the SD
    CheckError(r)

    ' this method relies on the SD file using contiguous sectors !!! Only the first 128 bytes of 512 bytes used
    sd_block_number := (drive_base[hdsk_drive+8]) + (hdsk_track * 32) + hdsk_sector
    err := sd.writeSDCard(sd_block_number, @disk_buff, 128)    'write a sector
'   CheckError(err)

    r := sd.stopSDcard                                  'stop the SD (free up pins, but keep cog running)
    CheckError(r)

    r := tbp2.active                                    'activate the bus
    CheckError(r)
'----------------------------------------------------------------------------------------------
#else
'----------------------------------------------------------------------------------------------
  'read the block in from Z80 sram
  r := tbp2.DoCmd("R", @disk_buff, hdsk_dma, 128)       'Virtual "DMA" the sector from DMA buffer address from SRAM
  CheckError(r)

  r := tbp2.idle                                        'give up the bus
  CheckError(r)

  r := sd.initSDcard(spiDO,spiClk,spiDI,spiCS)          'init the SD
  CheckError(r)

  ' this method relies on the SD file using contiguous sectors !!! Only the first 128 bytes of 512 bytes used
  sd_block := (drive_base[hdsk_drive+8]) + (hdsk_track * 32) + hdsk_sector
  err := sd.writeSDCard(sd_block, @disk_buff, 128)      'write a sector
' CheckError(err)

  r := sd.stopSDcard                                    'stop the SD (free up pins, but keep cog running)
  CheckError(r)

  r := tbp2.active                                      'activate the bus
  CheckError(r)
'----------------------------------------------------------------------------------------------
#endif
#endif

  'Virtual "DMA" the sector from DMA buffer address in Z80 RAM
#ifdef PropDemoBoard
  cpu.dma_read (hdsk_dma, @disk_buff + block_offset)
  err := \sd.writeblock(sd_block_number, @disk_buff)  'Write the block! (complete 512 bytes)
#elseifdef TriBladeProp
  'Cluso perhaps something here
#endif

  if err
    UART.str(string("Error hdisk write "))
    DisplayError(err)
    DisplayHDiskDTS
    io_data := $FF                                      'Return error code
  else
    io_data := $00                                      'Return OK.

  hdsk_last_command := hdskNull                         'Command completed


PRI in_hdskport_param
'Handle read from hdskport during disk parameters command
'Simply return the next param byte from a DPB.
  'UART.str(string("hdskparam "))
  'UART.dec(hdsk_reply_pos)
  'UART.str(string(" = "))
  io_data := hdsk_dpb[hdsk_reply_pos++]
  io_data := BYTE[hdsk_dpb_ptr++]
  hdsk_reply_pos++
  'UART.hex(io_data,2)
  'crlf
  if hdsk_reply_pos == 19
    hdsk_reply_pos := 0
    hdsk_last_command := hdskNull                   'Command completed
    UART.str(string("command completed "))
    crlf


PRI in_hdskport
'Handle reads from the hard disk interface port
'Handling is dispatched to an appropriate function depending on the disk operation currently in progress
  case hdsk_last_command
    hdskRead:
      in_hdskport_read
    hdskWrite:
      in_hdskport_write
    hdskParam:                                          'Current command is get disk params so just return successive parameter bytes from DPB.
      in_hdskport_param
    other:
      UART.str(string("HDSK ERROR 5"))
      crlf


PRI on_output
'Handle Z80 IO port writes
  case io_port
    ledport:                                            'An 8 bit parallel output with LEDs attached
      out_ledport
    simhport:                                           'SIMH control port
      out_simhport
    condata:                                            'Console data port
      out_condata
    constat:                                            'Console status port
      out_constat
    punstat:
      out_punstat                                       'Punch control port
    n8vem_uart_out:                                     '&H68
      out_n8vem_byte
    hdskport:
      out_hdskport                                      'SIMH style hard disk port
    other:
      UART.str(string("Write to unimplemented I/O Port &H"))
      UART.hex(io_port,2)                               'print the port number for debugging
      crlf
  io_command := 0                                       'Set flag ready for next io

PRI out_ledport
'Handle writes to the LED port
'No, SIMH never had this. I added it for the Propeller demo board.
  'outa[16..23] := io_data                              'Output the port data  (if re-enabled do #ifdef)
  UART.str(string("LED port = "))
  UART.hex(io_data,2)
  crlf


PRI out_simhport
'Handle writes to the SIMH simulator control port


PRI out_condata
'Handle writes to the console data output port
    UART.tx(io_data)


PRI out_constat
'Handle writes to the console command port
  if io_data == creset
    'UART.str(string("Console reset"))
    'crlf
  else
    UART.str(string("Unknown console status write $"))
    UART.hex(io_data,2)
    crlf


PRI out_punstat
'Handle writes to the punch device command port
  if io_data == creset
    'UART.str(string("Punch reset"))
    'crlf
  else
    UART.str(string("Unknown punch status write $"))
    UART.hex(io_data,2)
    crlf             


PRI out_n8vem_byte
  UART.tx(io_data)   ' send out the data to the port (for xmodem)


PRI hdsk_get_rw_params
  case hdsk_command_pos
    1:
      hdsk_drive := io_data                             '1st byte is the drive number
      hdsk_command_pos++
    2:
      hdsk_sector := io_data                            '2nd byte is the sector number
      hdsk_command_pos++
    3:
      hdsk_track := io_data                             '3rd byte is LOW of the track number
      hdsk_command_pos++
    4:
      hdsk_track := hdsk_track + (io_data << 8)         '4th byte is HIGH of the track number
      hdsk_command_pos++
    5:
      hdsk_dma := io_data                               '5th byte is LOW of the DMA address
      hdsk_command_pos++
    6:
      hdsk_dma := hdsk_dma + (io_data << 8)             '6th (and last) byte is HIGH of the DMA address
      hdsk_command_pos++
      hdsk_last_command := hdsk_current_command         'OUT's for current command completed
      hdsk_current_command := hdskNull
    other:
      UART.str(string("ERROR HDSK 1"))
      crlf
      repeat


PRI out_hdskport_param
'Handle writes to the hdskport when get params command is in progress
  case hdsk_command_pos
    1:
      hdsk_drive := io_data
      'UART.str(string("Hdsk drive = "))
      'UART.dec(hdsk_drive)
      'crlf
      hdsk_command_pos++
      hdsk_reply_pos := 0
      hdsk_dpb_ptr := @hdsk_dpb + (hdsk_drive * 19)     'Point to correct DPB for this drive
      hdsk_last_command := hdsk_current_command         'OUT's for current command completed
      hdsk_current_command := hdskNull
    other:
      UART.str(string("HDSK ERROR 4"))
      crlf
      repeat


PRI out_hdskport_null
'Handle writes to the hdskport when no command is in progress
'io_data should be a new command byte.
  case io_data
    hdskRead:
      'UART.str(string("hdsk read"))
      'crlf
      hdsk_current_command := hdskRead
      hdsk_command_pos := 1
    hdskWrite:
      'UART.str(string("hdsk write"))
      'crlf
      UART.tx(".")                                      'indicates writing as takes long time (looks like hang)
      hdsk_current_command := hdskWrite
      hdsk_command_pos := 1
    hdskParam:
      'UART.str(string("hdsk param"))
      'crlf
      hdsk_current_command := hdskParam
      hdsk_command_pos := 1
    hdskReset:
      'UART.str(string("hdsk reset"))
      'crlf
      hdsk_current_command := hdskNull
      hdsk_last_command := hdskNull
      hdsk_command_pos := 1
      hdsk_reply_pos := 0
    other:
      UART.str(string("HDSK ERROR 2"))
      crlf
      repeat


PRI out_hdskport
'Handle writes to the SIMH style hard disk port
'Handling is dispatched to an apropriate function depending
'on the command currently in progress.
'FIXME: RESET should be checked on every write no matter the command in progress.
  case hdsk_current_command
    hdskRead:
      hdsk_get_rw_params
    hdskWrite:
      hdsk_get_rw_params
    hdskParam:
      out_hdskport_param
    hdskReset:
      UART.str(string("HDSK This cannot happen!"))
      crlf
      repeat
    hdskNull:
      out_hdskport_null
    other:
      UART.str(string("HDSK ERROR 3"))
      crlf
      repeat


PRI DisplayHDiskDTS
  UART.str(string("D = "))
  UART.hex(hdsk_drive, 2)
  UART.str(string(" T = "))
  UART.hex(hdsk_track, 4)
  UART.str(string(" S = "))
  UART.hex(hdsk_sector, 2)
  UART.str(string(" DMA = "))
  UART.hex(hdsk_dma, 4)
  crlf


PRI print_regs
'Print the Z80 registers
  UART.str(string("PC="))
  UART.hex(pc_reg, 4)
  UART.tx(" ")
'  UART.hex(byte[ram_base + pc_reg], 2)

  UART.str(string(" AF="))
  UART.hex(af_reg, 4)
      
  UART.str(string(" BC="))
  UART.hex(bc_reg, 4)
      
  UART.str(string(" DE="))
  UART.hex(de_reg, 4)

  UART.str(string(" HL="))
  UART.hex(hl_reg, 4)
      
  UART.str(string(" IX="))
  UART.hex(ix_reg, 4)

  UART.str(string(" IY="))
  UART.hex(iy_reg, 4)

  UART.str(string(" SP="))
  UART.hex(sp_reg, 4)
  crlf


PRI crlf
  UART.tx(13)
  UART.tx(10)
